#include <pybind11/pybind11.h>
#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define BLOCK_SIZE 256
#define WARP_SIZE 32

// This kernel fuses ReLU and Group Normalization on the output of conv_transpose3d.
// It ensures memory coalescing by aligning global memory accesses using vectorized loads/stores (float4) and
// by loading the per-channel gamma and beta values into shared memory. Threads in a warp access consecutive
// memory locations, which minimizes memory transaction overhead.

// Kernel parameters:
//   data: pointer to tensor of shape [N, C, D, H, W]
//   gamma, beta: GroupNorm parameters (each of length C)
//   N, C, D, H, W: tensor dimensions
//   G: number of groups; channels are divided into G contiguous groups
//   eps: epsilon for numerical stability in variance computation

__global__ void fused_relu_groupnorm_coalesced_kernel(
    float* __restrict__ data,
    const float* __restrict__ gamma,
    const float* __restrict__ beta,
    int N, int C, int D, int H, int W,
    int G, float eps) {

    // Each block processes one (sample, group) pair
    int n = blockIdx.x;       // sample index
    int g = blockIdx.y;       // group index
    int channels_per_group = C / G;
    int spatial_size = D * H * W;
    int group_elems = channels_per_group * spatial_size;

    // Compute the starting offset for this group's data in the tensor
    int group_offset = n * (C * spatial_size) + g * channels_per_group * spatial_size;
    float* group_data = data + group_offset;

    // Allocate dynamic shared memory:
    // First part: two arrays for per-channel gamma and beta (size: channels_per_group each).
    // Second part: two arrays (of fixed size 32) for warp-level reduction of sum and sumsq.
    extern __shared__ float s_mem[];
    float* sh_gamma = s_mem;                           // size: channels_per_group
    float* sh_beta  = sh_gamma + channels_per_group;     // size: channels_per_group
    // Next 64 floats reserved for warp reduction (32 for sum, 32 for sumsq)
    float* s_sum    = sh_beta + channels_per_group;      // size: up to 32 floats
    float* s_sumsq  = s_sum + 32;                         // size: up to 32 floats

    // Load GroupNorm parameters for this group into shared memory for faster access
    if (threadIdx.x < channels_per_group) {
        sh_gamma[threadIdx.x] = gamma[g * channels_per_group + threadIdx.x];
        sh_beta[threadIdx.x]  = beta[g * channels_per_group + threadIdx.x];
    }
    __syncthreads();

    // First pass: Apply ReLU and compute the sum and sumofsquares over all elements in the group.
    // We use vectorized loads/stores (float4) to ensure that threads in a warp access consecutive memory locations.
    int total_vectors = group_elems / 4; // number of float4 elements
    int remainder = group_elems % 4;       // remaining elements that are not part of a full float4

    float local_sum = 0.0f;
    float local_sumsq = 0.0f;

    // Reinterpret the group's data as a float4 pointer for coalesced access
    float4* group_data4 = reinterpret_cast<float4*>(group_data);

    // Process the vectorized portion
    for (int i = threadIdx.x; i < total_vectors; i += blockDim.x) {
        float4 val4 = group_data4[i];
        // Apply ReLU element-wise
        float x0 = fmaxf(val4.x, 0.0f);
        float x1 = fmaxf(val4.y, 0.0f);
        float x2 = fmaxf(val4.z, 0.0f);
        float x3 = fmaxf(val4.w, 0.0f);
        
        // Store the activated values back
        val4.x = x0; val4.y = x1; val4.z = x2; val4.w = x3;
        group_data4[i] = val4;

        // Accumulate the sums and sums of squares
        local_sum   += (x0 + x1 + x2 + x3);
        local_sumsq += (x0*x0 + x1*x1 + x2*x2 + x3*x3);
    }

    // Process the remainder elements with scalar accesses
    int offset = total_vectors * 4;
    for (int i = threadIdx.x; i < remainder; i += blockDim.x) {
        int idx = offset + i;
        float val = group_data[idx];
        float relu_val = fmaxf(val, 0.0f);
        group_data[idx] = relu_val;
        local_sum += relu_val;
        local_sumsq += relu_val * relu_val;
    }

    // Warp-level reduction using shuffle intrinsics
    unsigned int mask = 0xffffffff;
    int lane = threadIdx.x & (WARP_SIZE - 1);
    for (int offset = WARP_SIZE/2; offset > 0; offset /= 2) {
        local_sum   += __shfl_down_sync(mask, local_sum, offset);
        local_sumsq += __shfl_down_sync(mask, local_sumsq, offset);
    }

    // Each warp’s lane 0 writes its result into shared memory
    int warp_id = threadIdx.x / WARP_SIZE;
    if (lane == 0) {
        s_sum[warp_id] = local_sum;
        s_sumsq[warp_id] = local_sumsq;
    }
    __syncthreads();

    // Thread 0 of the block performs the final reduction and computes group mean and inverse std.
    float group_mean, inv_std;
    if (threadIdx.x == 0) {
        int num_warps = (blockDim.x + WARP_SIZE - 1) / WARP_SIZE;
        float sum_total = 0.0f;
        float sumsq_total = 0.0f;
        for (int i = 0; i < num_warps; i++) {
            sum_total  += s_sum[i];
            sumsq_total += s_sumsq[i];
        }
        group_mean = sum_total / group_elems;
        float variance = sumsq_total / group_elems - group_mean * group_mean;
        inv_std = rsqrtf(variance + eps);
        // Store for broadcasting
        s_sum[0] = group_mean;
        s_sumsq[0] = inv_std;
    }
    __syncthreads();
    group_mean = s_sum[0];
    inv_std = s_sumsq[0];

    // Second pass: Normalize the data using the computed group statistics and the per-channel gamma and beta
    // Process the vectorized portion
    for (int i = threadIdx.x; i < total_vectors; i += blockDim.x) {
        float4 val4 = group_data4[i];
        int base_index = i * 4;  // linear index of the first element in this float4
        // Process each component of the float4
        #pragma unroll
        for (int j = 0; j < 4; j++) {
            int idx = base_index + j;            // overall index within the group
            int channel_idx = idx / spatial_size; // determine which channel this element belongs to
            float gamma_val = sh_gamma[channel_idx];
            float beta_val = sh_beta[channel_idx];
            float x = ((&val4.x)[j] - group_mean) * inv_std;
            (&val4.x)[j] = x * gamma_val + beta_val;
        }
        group_data4[i] = val4;
    }

    // Process any remaining elements
    for (int i = threadIdx.x; i < remainder; i += blockDim.x) {
        int idx = total_vectors * 4 + i;
        int channel_idx = idx / spatial_size;
        float x = (group_data[idx] - group_mean) * inv_std;
        group_data[idx] = x * sh_gamma[channel_idx] + sh_beta[channel_idx];
    }
}

// The forward function performs 3D transposed convolution using ATen and then launches the custom fused kernel
// to apply ReLU activation and Group Normalization with aligned, coalesced global memory accesses.

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor conv_transpose,
    torch::Tensor group_norm_weight,
    torch::Tensor group_norm_bias,
    int64_t groups,
    double eps) {

    auto y = at::conv_transpose3d(
         x,
         conv_transpose,
         /*bias=*/c10::nullopt,
         /*stride=*/{1, 1, 1},
         /*padding=*/{0, 0, 0},
         /*output_padding=*/{0, 0, 0},
         /*groups=*/1,
         /*dilation=*/{1, 1, 1}
    );

    int N = y.size(0);
    int C = y.size(1);
    int D = y.size(2);
    int H = y.size(3);
    int W = y.size(4);
    int G = groups;
    int channels_per_group = C / G;
    
    // Compute shared memory size:
    //  - sh_gamma and sh_beta: 2 * channels_per_group floats
    //  - Warp reduction arrays: 64 floats (32 for sum and 32 for sumsq)
    size_t shared_mem_size = (channels_per_group * 2 + 64) * sizeof(float);
    
    dim3 grid(N, G);
    dim3 block(BLOCK_SIZE);
    
    fused_relu_groupnorm_coalesced_kernel<<<grid, block, shared_mem_size>>>(
         y.data_ptr<float>(),
         group_norm_weight.data_ptr<float>(),
         group_norm_bias.data_ptr<float>(),
         N, C, D, H, W,
         G, static_cast<float>(eps)
    );
    
    cudaDeviceSynchronize();
    return y;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused ConvTranspose3D + ReLU + GroupNorm with coalesced global memory accesses (CUDA)");
}
